/*#io
Collector ioDoc(
			 docCopyright("Steve Dekorte", 2002)
			 docLicense("BSD revised")
			 docCategory("Core")
			 docDescription("""Contains methods related to Io's garbage collector. Io currently uses a incremental, non-moving, generational collector based on the tri-color (black/gray/white) algorithm with a write-barrier. 
<p>
Every N number of object allocs, the collector will walk some of the objects marked as gray, marking their connected white objects as gray and turning themselves black. Every M allocs, it will pause for a sweep where it makes sure all grays are marked black and frees all whites. 
<p>
If the sweepsPerGeneration is set to zero, it will immediately mark all blacks as white again and mark the root objects as gray. Otherwise, it will wait until the sweepsPerGeneration count is reached to do this. By adjusting the allocsPerSweep and sweepsPerGeneration appropriately, the collector can be tuned efficiently for various usage cases. Generally, the more objects in your heap, the larger you'll want this number.""")
*/

#include "IoCollector.h"
#include "IoNumber.h"

IoObject *IoCollector_proto(void *state)
{
    IoMethodTable methodTable[] = 
    { 
	{"debugOn", IoObject_collectorDebugOn},
	{"debugOff", IoObject_collectorDebugOff},
	{"collect", IoObject_collectGarbage},

	{"setAllocsPerSweep", IoObject_setAllocsPerSweep},
	{"allocsPerSweep", IoObject_allocsPerSweep},
	
	{"setSweepsPerGeneration", IoObject_setSweepsPerGeneration},
	{"sweepsPerGeneration", IoObject_sweepsPerGeneration},
	{0x0, 0x0},
    };
    
    IoObject *self = IoObject_new(state);
    IoObject_addMethodTable_(self, methodTable);
    return self;
}


IoObject *IoObject_collectGarbage(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("collect", "Runs garbage collector. Returns the number of items collected. ")
    */
    
    int count = IoState_collectGarbage(IOSTATE);
    //count += IoState_collectGarbage(IOSTATE);
    return IONUMBER(count); 
}

IoObject *IoObject_collectorDebugOn(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("debugOn", "Turn on printing of collector debugging messages. Returns self.")
    */
    
    IOSTATE->debugCollector = 1;
    return self; 
}

IoObject *IoObject_collectorDebugOff(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("debugOff", "Turn off printing of collector debugging messages. Returns self.")
    */
    
    IOSTATE->debugCollector = 0;
    return self; 
}

IoObject *IoObject_setAllocsPerSweep(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("setAllocsPerSweep(aNumber)", "Sets the number of object allocations per sweep. Returns self.")
    */
    
    int n = IoMessage_locals_intArgAt_(m, locals, 0);
    
    IOSTATE->allocsPerSweep = n;
    return self; 
}

IoObject *IoObject_allocsPerSweep(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("allocsPerSweep", 
    "Return the number of allocations per collector sweep.")
    */
    
    return IONUMBER(IOSTATE->allocsPerSweep);
}


IoObject *IoObject_setSweepsPerGeneration(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("setSweepsPerGeneration(aNumber)", 
    "Sets the number of sweeps before black marks are turned to white. Returns self.")
    */
    
    int n = IoMessage_locals_intArgAt_(m, locals, 0);
    
    IOSTATE->sweepsPerGeneration = n;
    return self; 
}

IoObject *IoObject_sweepsPerGeneration(IoObject *self, IoObject *locals, IoMessage *m)
{ 
    /*#io
    docSlot("sweepsPerGeneration", "Return the number of sweeps per generation.")
    */
    
    return IONUMBER(IOSTATE->sweepsPerGeneration);
}
