/*
 docCopyright("Steve Dekorte", 2002)
 docLicense("BSD revised")
 */

#define IOSTATE_C 1
#include "IoState.h"
#undef IOSTATE_C 

#include "IoObject.h"
#include "IoActivationContext.h"
#include "IoCoroutine.h"
#include "IoSeq.h"
#include "IoNumber.h"
#include "IoCFunction.h" 
#include "IoBlock.h"
#include "IoList.h"
#include "IoMap.h"
#include "IoFile.h"
#include "IoDate.h"
#include "IoDuration.h"
#include "IoSeq.h"
#include "IoMessage_parser.h"
#include "IoDynLib.h"
#include "IoWeakLink.h"

#include "IoSystem.h"
#include "IoCompiler.h"
#include "IoDebugger.h"
#include "IoCollector.h"
#include "IoScheduler.h"
#include "IoRandom.h"
#include "IoSandbox.h"
//#include "IoPObject.h"
#include "IoDirectory.h"

#include <limits.h> 
#include <stdlib.h>

void IoVMCodeInit(IoState *self, IoObject *context);

void IoState_setupQuickAccessSymbols(IoState *self)
{
	self->activateString     = IoState_retain_(self, SIOSYMBOL("activate"));
	self->cloneString	     = IoState_retain_(self, SIOSYMBOL("clone"));
	self->CompilerString     = IoState_retain_(self, SIOSYMBOL("Compiler"));
	self->debugString	     = IoState_retain_(self, SIOSYMBOL("debug"));  
	self->debugParserString  = IoState_retain_(self, SIOSYMBOL("debugParser"));  
	self->equalEqualString   = IoState_retain_(self, SIOSYMBOL("=="));  
	self->forwardString	     = IoState_retain_(self, SIOSYMBOL("forward"));
	self->initString	     = IoState_retain_(self, SIOSYMBOL("init"));
	self->objectString	     = IoState_retain_(self, SIOSYMBOL("Object"));  
	self->parentString	     = IoState_retain_(self, SIOSYMBOL("parent"));
	self->selfString	     = IoState_retain_(self, SIOSYMBOL("self"));
	self->setSlotString	     = IoState_retain_(self, SIOSYMBOL("setSlot"));
	self->updateSlotString   = IoState_retain_(self, SIOSYMBOL("updateSlot"));
	self->callString         = IoState_retain_(self, SIOSYMBOL("call"));  
}

void IoState_setupCollector(IoState *self)
{
	self->valueToWeakLink = Hash_new();
	//self->collectedWeakLinks = Stack_new();
	
	self->freeGroup  = IoObjectGroup_newWithColor_(self, IOMARK_INITIAL_FREE);
	self->blackGroup = IoObjectGroup_newWithColor_(self, IOMARK_INITIAL_BLACK);
	self->grayGroup  = IoObjectGroup_newWithColor_(self, IOMARK_INITIAL_GRAY);
	self->whiteGroup = IoObjectGroup_newWithColor_(self, IOMARK_INITIAL_WHITE);
	
	IoState_updateCachedCollectorColors(self);
	
	self->recycledObjects = List_new();
	self->recycledObjectsSansSlots = List_new();
	self->newValueCount = 0;
}

IoState *IoState_new(void)
{
	IoCFunction *cFunctionProto;
	IoSeq *seqProto;
	IoState *self = (IoState *)calloc(1, sizeof(IoState));
	
	self->mainArgs = MainArgs_new();
	self->primitiveProtos = Hash_new();
	
	IoState_setupCollector(self);
	IoState_pushCollectorPause(self);
	
	// Sandbox 
	
	self->messageCount = 0;
	self->messageCountLimit = 0;
	self->endTime = 0;
	
	self->sdbm = SkipDBM_new();
	self->symbols = SkipDBM_rootSkipDB(self->sdbm);
	//SkipDB_setCompareFunction_(self->symbols, (SkipDBCompareFunc *)ioStateStringCmp);
	//SkipDB_objectMarkFunc_(self->symbols, (SkipDBObjectMarkFunc *)IoObject_makeGrayIfWhite);
	
	self->retainedValues = List_new();
	
	self->scheduler = Scheduler_new(self);
	Scheduler_callbackContext_(self->scheduler, self);
	Scheduler_updateCallback_(self->scheduler, 
						 (SchedulerUpdateCallback *)IoState_schedulerUpdate);
	
	
	/* 
Problem:
	 - there are some interdependencies here:
	 - creating instances requires a retain stack
	 - we need a Coroutine to use for our retainStack
	 - defining any primitive methods requires Strings and CFunctions
	 
	 Solution:
	 - create Object, CFunction and String protos sans methods. 
	 - then add methods to Object, CFunction and String
	 */
	
	self->currentIoStack = Stack_new(); // temp retain stack until coro is up
	
	self->objectProto = IoObject_proto(self); // need to do this first, so we have a retain stack 
	
	self->mainCoroutine = IoCoroutine_proto(self);
	Stack_free(self->currentIoStack);
	self->currentIoStack = 0x0;
	
	IoState_setCurrentCoroutine_(self, self->mainCoroutine);
	IoState_retain_(self, self->mainCoroutine);
	
	IoState_addValue_(self, self->objectProto); // to put objectProto into an ObjectGroup
	
	seqProto = IoSeq_proto(self);
	
	IoState_setupQuickAccessSymbols(self);
	
	IoObject_rawSetProto_(seqProto, self->objectProto); 
	
	cFunctionProto = IoCFunction_proto(self);
	self->localsUpdateSlotCFunc = IoState_retain_(self, 
										 IoCFunction_newWithFunctionPointer_tag_name_(self, IoObject_localsUpdateSlot, 0x0, "localsUpdate"));
	
	IoSeq_protoFinish(seqProto);
	IoObject_protoFinish(self);
	IoCFunction_protoFinish(self);
	IoCoroutine_protoFinish(self->mainCoroutine);
	
	self->setSlotBlock  = IoState_retain_(self, 
								   IoObject_getSlot_(self->objectProto, SIOSYMBOL("setSlot")));  
	
	// setup lobby
	
	{
		IoObject *objectProto = self->objectProto; 
		IoObject *protos = IOCLONE(objectProto);
		IoObject *vm     = IOCLONE(objectProto);
		
		self->lobby      = IOCLONE(objectProto);
		IoState_retain_(self, self->lobby);
		
		//printf("State objectProto = %p\n", (void *)objectProto);
		//IoState_retain_(self, objectProto); //temp
		
		// setup namespace
		
		IoObject_setSlot_to_(self->lobby, SIOSYMBOL("Lobby"), self->lobby);
		IoObject_setSlot_to_(self->lobby, SIOSYMBOL("Protos"), protos);  
		IoObject_setSlot_to_(protos, SIOSYMBOL("IoVM"), vm); 
		
		IoObject_setSlot_to_(vm, SIOSYMBOL("Compiler"),  IoCompiler_proto(self));
		IoObject_setSlot_to_(vm, SIOSYMBOL("Collector"), IoCollector_proto(self));
		IoObject_setSlot_to_(vm, SIOSYMBOL("Scheduler"), IoScheduler_proto(self));
		IoObject_setSlot_to_(vm, SIOSYMBOL("Exception"), IOCLONE(objectProto));
		
		// setup proto chain
		
		IoObject_rawSetProto_(objectProto, self->lobby); 
		IoObject_rawSetProto_(self->lobby, protos); 
		IoObject_rawSetProto_(protos, vm); 
		
		// add protos to namespace
		
		IoObject_setSlot_to_(vm, SIOSYMBOL("Object"), objectProto);
		IoObject_setSlot_to_(vm, SIOSYMBOL("Sequence"), seqProto);
		IoObject_setSlot_to_(vm, SIOSYMBOL("Number"), IoNumber_proto(self)); 
		IoObject_setSlot_to_(vm, SIOSYMBOL("Random"), IoRandom_proto(self)); 
		IoObject_setSlot_to_(vm, SIOSYMBOL("System"), IoSystem_proto(self));
		
		// nil
		self->ioNil = IOCLONE(objectProto);
		IoObject_setSlot_to_(vm, SIOSYMBOL("Nil"), self->ioNil);
		IoObject_setSlot_to_(vm, SIOSYMBOL("nil"), self->ioNil);
		IoObject_setSlot_to_(vm, SIOSYMBOL("Message"), IoMessage_proto(self));
		IoObject_setSlot_to_(vm, SIOSYMBOL("ActivationContext"),  IoActivationContext_proto(self));
		
		self->nilMessage  = IoMessage_newWithName_(self, SIOSYMBOL("nil"));
		IoMessage_cachedResult_(self->nilMessage, self->ioNil);
		IoState_retain_(self, self->nilMessage);
		
		// true 
		self->ioTrue = IoObject_new(self);
		IoObject_setSlot_to_(vm, SIOSYMBOL("true"), self->ioTrue);
		IoObject_setSlot_to_(self->ioTrue, SIOSYMBOL("type"), SIOSYMBOL("true"));
		IoState_retain_(self, self->ioTrue);
		
		// false
		self->ioFalse = IoObject_new(self);
		IoObject_setSlot_to_(vm, SIOSYMBOL("false"), self->ioFalse);
		IoObject_setSlot_to_(self->ioFalse, SIOSYMBOL("type"), SIOSYMBOL("false"));
		IoState_retain_(self, self->ioFalse);
		
		// cached messages
		
		self->printMessage  = IoMessage_newWithName_(self, SIOSYMBOL("print"));
		IoState_retain_(self, self->printMessage);
		
		self->initMessage   = IoMessage_newWithName_(self, SIOSYMBOL("init"));
		IoState_retain_(self, self->initMessage);
		
		self->compareMessage = IoMessage_newWithName_(self, SIOSYMBOL("compare"));
		IoState_retain_(self, self->compareMessage);
		
		self->runMessage = IoMessage_newWithName_(self, SIOSYMBOL("run"));
		IoState_retain_(self, self->runMessage);
		
		self->mainMessage = IoMessage_newWithName_(self, SIOSYMBOL("main"));
		IoState_retain_(self, self->mainMessage);
		
		{
			IoObject *debugger = IoDebugger_proto(self);
			IoState_retain_(self, debugger);
			IoObject_setSlot_to_(vm, SIOSYMBOL("Debugger"), debugger);
			self->debugger = debugger;
			
			self->vmWillSendMessage  = IoMessage_newWithName_(self, SIOSYMBOL("vmWillSendMessage"));
			IoMessage_cachedResult_(self->nilMessage, self->ioNil);
			IoState_retain_(self, self->vmWillSendMessage);
		}
		
		IoObject_setSlot_to_(vm, SIOSYMBOL("Block"),      IoBlock_proto(self));
		IoObject_setSlot_to_(vm, SIOSYMBOL("List"),       IoList_proto(self));
		IoObject_setSlot_to_(vm, SIOSYMBOL("Map"),        IoMap_proto(self));
		IoObject_setSlot_to_(vm, SIOSYMBOL("Coroutine"),  self->mainCoroutine);
		IoObject_setSlot_to_(vm, SIOSYMBOL("File"),       IoFile_proto(self));
		IoObject_setSlot_to_(vm, SIOSYMBOL("Directory"),  IoDirectory_proto(self));
		IoObject_setSlot_to_(vm, SIOSYMBOL("Date"),       IoDate_proto(self));
		IoObject_setSlot_to_(vm, SIOSYMBOL("Duration"),   IoDuration_proto(self));
		IoObject_setSlot_to_(vm, SIOSYMBOL("WeakLink"),   IoWeakLink_proto(self));
		IoObject_setSlot_to_(vm, SIOSYMBOL("Sandbox"),    IoSandbox_proto(self));
		
#if !defined(__SYMBIAN32__)
		IoObject_setSlot_to_(vm, SIOSYMBOL("DynLib"),     IoDynLib_proto(self));
#endif
		
		self->store = IoStore_proto(self);		
		IoObject_setSlot_to_(vm, SIOSYMBOL("Store"),      self->store);
		//IoObject_setSlot_to_(vm, SIOSYMBOL("PObject"),  IoPObject_proto(self))
		IoObject_setSlot_to_(vm, SIOSYMBOL("CFunction"),  cFunctionProto);
		
		self->localsProto = IoState_retain_(self, IoObject_localsProto(self));
		IoObject_setSlot_to_(vm, SIOSYMBOL("Locals"),  self->localsProto);
		
		self->stopStatus = MESSAGE_STOP_STATUS_NORMAL;
		self->returnValue = (void *)0x0;
		
		IoState_gcInitPhase(self);
		self->newValueCountToUpdate = 10; 
		self->allocsPerSweep  = STATE_DEFAULT_COLECT_AFTER;
		
		IoCoroutine_rawSetLabel_(self->mainCoroutine, SIOSYMBOL("main"));
		IoState_clearRetainStack(self);
		IoState_popCollectorPause(self);
		//IoState_gcSweepPhase(self);
		
		IoVMCodeInit(self, vm);
		IoState_clearRetainStack(self);
	}
	
	return self;
}

IoObject *IoObject_initBindings(IoObject *self, IoObject *locals, IoMessage *m)
{
	IOSTATE->bindingsInitCallback(IOSTATE, self);
	return self;
}

void IoState_init(IoState *self)
{
	if (self->bindingsInitCallback)
	{
		//IoCFunction *func;
		IoObject *vm = IoObject_getSlot_(self->lobby, SIOSYMBOL("IoVM"));
		
		IoState_pushCollectorPause(self);
		/*
		func = IoCFunction_newWithFunctionPointer_tag_name_(self, 
												  (IoUserFunction *)IoObject_initBindings, 
												  0x0, 
												  "[IoState]");
		
		IoObject_setSlot_to_(vm, SIOSYMBOL("initBindings"), func);
		
		IoState_withinCatchPerform(self, vm, vm, 
							  IoMessage_newWithName_(self, SIOSYMBOL("initBindings")));
		*/
		self->bindingsInitCallback(self, vm);

		IoState_popCollectorPause(self);
		IoState_clearRetainStack(self);
	}
}

void IoState_registerProtoWithFunc_(IoState *self, IoObject *proto, IoStateProtoFunc *func)
{ 
	if (Hash_at_(self->primitiveProtos, (void *)func))
	{
		IoState_fatalError_(self, "IoState_registerProtoWithFunc_() Error: attempt to add the same proto twice");
	}
	
	IoState_retain_(self, proto);
	Hash_at_put_(self->primitiveProtos, (void *)func, proto); 
	//printf("registered %s\n", IoObject_name(proto));
}

IoObject *IoState_protoWithName_(IoState *self, const char *name)
{
	IoObject *proto = Hash_firstValue(self->primitiveProtos);
	
	while (proto)
	{
		if (!strcmp(IoObject_name(proto), name)) 
		{
			return proto;
		}
		
		proto = Hash_nextValue(self->primitiveProtos);
	}
	
	return 0x0;
}

void IoObject_addProtoToCollector(IoObject *proto)
{
	IoState *state = proto->tag->state;
	
	if (strcmp("Object", IoObject_name(proto)) == 0) 
	{
		return;
	}
	
	// printf("addProtoToCollector(%s)\n", IoObject_name(proto));
	
	if (!IoState_hasValue_(state, proto)) 
	{
		IoState_addValue_(state, proto);
	}
	
	Hash_removeValue_(state->primitiveProtos, proto);
	IoState_collectGarbage(state);
}

List *IoState_tagList(IoState *self) // caller must free returned List 
{
	List *tags = List_new();
	void *k = Hash_firstKey(self->primitiveProtos);
	
	while (k)
	{
		IoObject *proto = Hash_at_(self->primitiveProtos, k);
		List_append_(tags, proto->tag);
		k = Hash_nextKey(self->primitiveProtos);
	}
	
	return tags;
}

void IoState_free(IoState *self)
{
	List *tags = IoState_tagList(self);
	// well need to free the tags after freeing their protos 
	
	self->isCleaningUp = 1;
	
	MainArgs_free(self->mainArgs);
	self->mainArgs = 0x0;
	
	//List_empty(self->coroutines);
	// need to abort all coros
	// if (self->currentCoroutine) IoState_retain_(self, self->currentCoroutine); 
	List_empty(self->retainedValues);
	
	{
		IoObject *v;
		
		while ((v = List_pop(self->recycledObjects))) 
		{
			IoObject_dealloc(v);
		}
		
		List_free(self->recycledObjects);
		self->recycledObjects = 0x0;
		
		while ((v = List_pop(self->recycledObjectsSansSlots))) 
		{
			IoObject_dealloc(v);
		}
		
		List_free(self->recycledObjectsSansSlots);
		self->recycledObjectsSansSlots = 0x0;
	}
	
	self->lobby = 0x0;
	IoState_collectGarbage(self);
	IoState_collectGarbage(self); // 2nd time to catch any non-white objects from last collect */
	
	//Hash_do_(self->primitiveProtos, (HashDoCallback *)addProtoToCollector);
	
	while (Hash_firstValue(self->primitiveProtos) && Hash_count(self->primitiveProtos) > 1)
	{
		IoObject_addProtoToCollector(Hash_firstValue(self->primitiveProtos));
	}
	
	//IoState_collectGarbage(self);  // final collection 
	
	SkipDBM_free(self->sdbm);
	self->sdbm = 0x0;
	self->symbols = 0x0;
	
	List_free(self->retainedValues);
	self->retainedValues = 0x0;
	
	// the protos that these tags were alloced 
	// in were already collected so we free them directly 
	
	List_do_(tags, (ListDoCallback *)IoTag_free); 
	Hash_free(self->primitiveProtos);
	self->primitiveProtos = 0x0;
		
	Hash_free(self->valueToWeakLink);
	self->valueToWeakLink = 0x0;
	
	Scheduler_free(self->scheduler);
	self->scheduler = 0x0;
	
	//Stack_free(self->collectedWeakLinks); 
	IoObjectGroup_free(self->freeGroup);
	self->freeGroup = 0x0;
	IoObjectGroup_free(self->blackGroup);
	self->blackGroup = 0x0;
	IoObjectGroup_free(self->grayGroup);
	self->grayGroup = 0x0;
	IoObjectGroup_free(self->whiteGroup);
	self->whiteGroup = 0x0;
	
	//IoStore_free(self->store);
	
	List_free(tags);
	free(self);
}

IoObject *IoState_lobby(IoState *self) 
{ 
	return self->lobby; 
}

void IoState_setLobby_(IoState *self, IoObject *obj) 
{ 
	self->lobby = obj;
}

IoObject *IoState_protoWithInitFunction_(IoState *self, IoStateProtoFunc *func)
{ 
	IoObject *proto = Hash_at_(self->primitiveProtos, (void *)func); 
	
	if (!proto)
	{
		IoState_fatalError_(self, "IoState_protoWithInitFunction_() Error: missing proto");
	}
	
	return proto;
}


// command line ------------------------------------------------

void IoState_argc_argv_(IoState *self, int argc, const char *argv[])
{
	IoList *args = IoList_new(self);
	int i;
	
	for (i = 1; i < argc; i ++)
	{
		IoList_rawAppend_(args, SIOSYMBOL(argv[i]));
	}
	
	IoObject_setSlot_to_(self->lobby, SIOSYMBOL("args"), args);
	
	MainArgs_argc_argv_(self->mainArgs, argc, argv); 
}

// store -------------------------------------------------------

IoStore *IoState_store(IoState *self)
{ 
	return self->store; 
}

IoObject *IoState_rawOn_doCString_withLabel_(IoState *self, 
									IoObject *target,
									const char *s,
									const char *label)
{
	IoMessage *m = IoMessage_newFromText_label_(self, s, label);
	return IoMessage_locals_performOn_(m, self->lobby, self->lobby);
}

void IoState_rawPrompt(IoState *self)
{
	int max = 1024 * 16;
	char *s = malloc(max);
	IoObject *result;
	
	for (;;)
	{
		fputs("Io> ", stdout);
		fflush(stdout);
		fgets(s, max, stdin);
		
		if (feof(stdin)) 
		{
			break;
		}
		
		//result = IoState_on_doCString_withLabel_(self, self->lobby, s, "IoState_rawPrompt()");
		result = IoState_rawOn_doCString_withLabel_(self, self->lobby, s, "IoState_rawPrompt()");
		
		//IoState_gcSweepPhase(self);
		fputs("==> ", stdout);
		IoObject_print(result);
		fputs("\n", stdout);
	}
	
	free(s);
}

void IoState_runCLI(IoState *self)
{
	//IoState_rawPrompt(self);
	IoState_on_doCString_withLabel_(self, self->lobby, "CLI run", "IoState_runCLI()");
}
