/*#io
System ioDoc(
		   docCopyright("Steve Dekorte", 2002)
		   docLicense("BSD revised")
		   docObject("System")
		   docDescription("Contains methods related to the IoVM.")
		   docCategory("Core")
		   */

#if defined(unix) || defined(__APPLE__) || defined(__NetBSD__)
#include <sys/utsname.h>
#endif

#include "IoSystem.h"
#include "IoNumber.h"
#include "IoMessage_parser.h"

#if defined(__CYGWIN__) || defined(_WIN32)
#include <windows.h>
#endif

IoObject *IoSystem_proto(void *state)
{
	IoMethodTable methodTable[] = { 
	{"errno", IoObject_errnoDescription},
	{"exit", IoObject_exit},
	{"getenv", IoObject_getenv},
	{"system", IoObject_system},
	{"memorySizeOfState", IoObject_memorySizeOfState},
	{"compactState", IoObject_compactState},
	{"platform", IoObject_platform},
	{0x0, 0x0},
	};
	
	IoObject *self = IoObject_new(state);
	IoObject_addMethodTable_(self, methodTable);
	
	IoObject_setSlot_to_(self, IOSYMBOL("version"), IONUMBER(IO_VERSION_NUMBER));
	IoObject_setSlot_to_(self, IOSYMBOL("distribution"), IOSYMBOL("Io"));
	
	return self;
}

/*
IoObject *IoObject_errno(IoObject *self, IoObject *locals, IoMessage *m)
{
	return IONUMBER(errno);
}
*/

#include <stdio.h>
#include <errno.h>

IoObject *IoObject_errnoDescription(IoObject *self, IoObject *locals, IoMessage *m)
{
	return errno ? IOSYMBOL(strerror(errno)) : IONIL(self);
}

IoObject *IoObject_exit(IoObject *self, IoObject *locals, IoMessage *m)
{ 
	/*#io
	docSlot("exit(optionalReturnCodeNumber)", 
		   "Shutdown the IoState (free all objects) and return 
control to the calling program (if any). ")
	*/
	
	int returnCode = 0;
	
	if (IoMessage_argCount(m))
	{
		returnCode = IoMessage_locals_intArgAt_(m, locals, 0);
	}
	
	IoState_exit(IOSTATE, returnCode); 
	return self; 
}

IoObject *IoObject_getenv(IoObject *self, IoObject *locals, IoMessage *m)
{ 
	/*#io
	docSlot("getenv(nameString)", 
		   "Returns a string with the value of the environment 
variable whose name is specified by nameString.") 
	*/
	
	IoSymbol *key = IoMessage_locals_symbolArgAt_(m, locals, 0);
	char *s = getenv(CSTRING(key));
	
	if (!s) 
	{
		return ((IoState *)IOSTATE)->ioNil;
	}
	
	return IoState_symbolWithCString_(IOSTATE, s);
}

IoObject *IoObject_system(IoObject *self, IoObject *locals, IoMessage *m)
{
	/*#io
	docSlot("system(aString)", 
		   "Makes a system call and returns a Number for the return value.")
	*/
	
	IoSymbol *s = IoMessage_locals_symbolArgAt_(m, locals, 0);
	int result = system(CSTRING(s));
	//printf("system result = %i", result);
	return IONUMBER((double)result);
}

IoObject *IoObject_memorySizeOfState(IoObject *self, IoObject *locals, IoMessage *m)
{ 
	/*
	 docSlot("memorySizeOfState", 
		    "Returns the number of bytes in the IoState 
	 (this may not include memory allocated by C libraries).")
	 */
	
	return IONUMBER(IoState_memorySize(IOSTATE)); 
}

IoObject *IoObject_compactState(IoObject *self, IoObject *locals, IoMessage *m)
{ 
	/*
	 docSlot("compactState", 
		    "Attempt to compact the memory of the IoState if possible.")
	 */
	
	IoState_compact(IOSTATE); 
	return self; 
}

/*
 
 IoObject *IoObject_setenv(IoObject *self, IoObject *locals, IoMessage *m)
 { 
      // setenv() takes different args in different implementations
	 IoSymbol *key = IoMessage_locals_symbolArgAt_(m, locals, 0);
	 IoSymbol *value = IoMessage_locals_symbolArgAt_(m, locals, 1);
	 setenv(CSTRING(key), CSTRING(value), 1); 
	 return self;
 }
 */

IoObject *IoObject_platform(IoObject *self, IoObject *locals, IoMessage *m)
{
	char *platform = "Unknown";
#if defined(__CYGWIN__) || defined(_WIN32)
	OSVERSIONINFO os;
	
	os.dwOSVersionInfoSize = sizeof( OSVERSIONINFO );
	GetVersionEx(&os);
	
	switch(os.dwPlatformId) 
	{
		case VER_PLATFORM_WIN32_WINDOWS:
			platform = "Windows 9X"; 
			break;
			
		case VER_PLATFORM_WIN32_NT:
			platform = "Windows NT"; 
			break;
	}
	
#elif defined(unix) || defined(__APPLE__) || defined(__NetBSD__)
	/* Why Apple and NetBSD don't define 'unix' I'll never know. */
	struct utsname os;
	int ret = uname(&os);
	
	if (ret == 0) 
	{
		platform = os.sysname;
	}
#endif
	
	return IoState_symbolWithCString_(self->state, platform);
}

