/*#io
WeakLink ioDoc(
               docCopyright("Steve Dekorte", 2002)
               docLicense("BSD revised")
			docCategory("Core")
               docDescription("""A WeakLink is a primitive that can hold a reference to an object without preventing the garbage collector from collecting it. The link reference is set with the setLink() method. After the garbage collector collects an object, it informs any WeakLink objects whose link value pointed to that object. If a WeakLink has a delegate set, it will send the delagate a collectedWeakLink() message with the WeakLink as an argument.
<p>
The identifier value of a WeakLink is usefull for storing miscellaneous information about the link if needed.
<p>

<b>What are weak links useful for?</b>
<p>

In general, they're handy when you want to monitor or cache an object without owning it. Examples:
<p>

<b>Distributed Objects example</b>
<p>

The original motivation for adding weaklinks was to deal with garbage collection in a distributed object system. In such a system, it can be usefull to keep a list of proxies that have been created to represent remote objects, but you don't want this list to cause those objects to be forever retained.
<p>

<b>Database Caching example</b>
<p>

In a persistance layer to a database it's often desirable to have separate queries return the same row objects if some other part of the application has already fetched them. This involves a cache, and weak links allow the cache to only keep database items that are still referenced by other parts of the application.
<p>

<b>Implementation notes</b>
<p>

When a WeakLink's link is set, it sets a weak link bit in the mark byte of the object being monitored. When the garbage collector collects an object, it checks this bit and if it's marked, looks in it's list of WeakLinks and sends a collected message to each WeakLink that references that object.
""")
               */

#include "IoWeakLink.h"
#include "IoSeq.h"
#include "IoState.h"
#include "IoObject.h"
#include "IoNumber.h"

#define DATA(self) ((IoWeakLinkData *)self->data)

IoTag *IoWeakLink_tag(void *state)
{
	IoTag *tag = IoTag_newWithName_("WeakLink");
	tag->state = state;
	tag->cloneFunc   = (TagCloneFunc *)IoWeakLink_rawClone;
	/*tag->compareFunc = (TagCompareFunc *)IoObject_compare;*/
	tag->markFunc    = (TagMarkFunc *)IoWeakLink_mark;
	tag->freeFunc    = (TagFreeFunc *)IoWeakLink_free;
	tag->writeToStoreOnStreamFunc = (TagWriteToStoreOnStreamFunc *)IoWeakLink_writeToStore_stream_;
	tag->readFromStoreOnStreamFunc = (TagReadFromStoreOnStreamFunc *)IoWeakLink_readFromStore_stream_;
	return tag;
}

void IoWeakLink_writeToStore_stream_(IoWeakLink *self, IoStore *store, BStream *stream)
{
	BStream_writeTaggedInt32_(stream, IoStore_pidForObject_(store, DATA(self)->link));
	BStream_writeTaggedInt32_(stream, IoStore_pidForObject_(store, DATA(self)->delegate));
	BStream_writeTaggedInt32_(stream, IoStore_pidForObject_(store, DATA(self)->identifier));
}

void IoWeakLink_readFromStore_stream_(IoWeakLink *self, IoStore *store, BStream *stream)
{
	IoObject *link = IoStore_objectWithPid_(store, BStream_readTaggedInt32(stream));
	IoWeakLink_justSetLink(self, link);
	DATA(self)->delegate = IoStore_objectWithPid_(store, BStream_readTaggedInt32(stream));
	DATA(self)->identifier = IoStore_objectWithPid_(store, BStream_readTaggedInt32(stream));
}

IoObject *IoWeakLink_proto(void *state) 
{ 
	IoMethodTable methodTable[] = {
	{"setLink", IoWeakLink_setLink},
	{"link", IoWeakLink_link},
	{"setDelegate", IoWeakLink_setDelegate},
	{"delegate", IoWeakLink_delegate},
	{"setIdentifier", IoWeakLink_setIdentifier},
	{"identifier", IoWeakLink_identifier},
	{NULL, NULL},
	};
	
	IoObject *self = IoObject_new(state);
	
	self->data = calloc(1, sizeof(IoWeakLinkData));
	self->tag = IoWeakLink_tag(state);
	DATA(self)->link       = IONIL(self);
	DATA(self)->delegate   = IONIL(self);
	DATA(self)->identifier = IONIL(self);
	IoState_registerProtoWithFunc_((IoState *)state, self, IoWeakLink_proto);
	
	IoState_addWeakLink_(IOSTATE, self);
	
	IoObject_addMethodTable_(self, methodTable);
	return self;
}

IoObject *IoWeakLink_rawClone(IoObject *proto) 
{ 
	IoObject *self = IoObject_rawClonePrimitive(proto);
	self->data = calloc(1, sizeof(IoWeakLinkData));
	
	DATA(self)->delegate   = DATA(proto)->delegate;
	DATA(self)->identifier = DATA(proto)->identifier;
	/*
	 if (DATA(proto)->link) 
	 IoWeakLink_justSetLink(self, DATA(proto)->link);
	 else
	 */
	DATA(self)->link = IONIL(proto);
	return self;
}

IoObject *IoWeakLink_new(void *state)
{ 
	IoObject *proto = IoState_protoWithInitFunction_((IoState *)state, IoWeakLink_proto);
	return IOCLONE(proto);
}

void IoWeakLink_free(IoObject *self)
{
	if (NULL == self->data) 
	{ 
		return; 
	}
	
	IoWeakLink_unlink(self);
	IoState_removeWeakLink_(IOSTATE, self);
	free(self->data);
	self->data = 0x0;
}

IoObject *IoWeakLink_newWithValue_(void *state, IoObject *v)
{
	IoObject *self = IoWeakLink_new(state);
	DATA(self)->link = v;
	return self;
}

void IoWeakLink_mark(IoObject *self)
{ 
	IoObject_makeGrayIfWhite(DATA(self)->delegate);
	IoObject_makeGrayIfWhite(DATA(self)->identifier);
	// we intentionaly dont mark link so it will get collected despited our reference 
}

IoObject *IoWeakLink_rawLink(IoWeakLink *self)
{ 
	return DATA(self)->link; 
}

IoObject *IoWeakLink_rawNextWeakLink(IoWeakLink *self)
{ 
	return DATA(self)->nextWeakLink; 
}

// special weaklink methods ----------------------------- 

void IoWeakLink_appendWeakLink_(IoObject *self, IoObject *other)
{
	if (DATA(self)->nextWeakLink) 
	{ 
		DATA(other)->nextWeakLink = DATA(self)->nextWeakLink;
		DATA(other)->previousWeakLink = self;
	}
	
	DATA(self)->nextWeakLink = other;
}

void IoWeakLink_unlink(IoObject *self) 
{
	if (DATA(self)->nextWeakLink) 
	{
		DATA(DATA(self)->nextWeakLink)->previousWeakLink = DATA(self)->previousWeakLink;
	}
	
	if (DATA(self)->previousWeakLink) 
	{
		DATA(DATA(self)->previousWeakLink)->nextWeakLink = DATA(self)->nextWeakLink;
	}
}

void IoWeakLink_collected(IoObject *self) 
{
	IoState *state = IOSTATE;
	// called by IoState when link is collected 
	// link may already be freed 
	DATA(self)->link = IONIL(self);
	
	if (!state->isCleaningUp)
	{
		// need to find a more efficient way to do this 
		IoMessage *m = IoMessage_newWithName_(state, IOSYMBOL("collectedWeakLink"));
		IoMessage_addArg_(m, IoMessage_newWithName_returnsValue_(state, IOSYMBOL("weaklink"), self));
		IoMessage_locals_performOn_(m, state->lobby, DATA(self)->delegate);
	}
	
	if (DATA(self)->nextWeakLink) 
	{
		IoWeakLink_collected(DATA(self)->nextWeakLink);
	}
}

void IoWeakLink_unlinkItem_(IoObject *self, IoObject *other)
{
	if (self == other)
	{
		IoWeakLink_unlink(self);
	}
	else
	{
		IoWeakLink_unlinkItem_(self, other);
	}
}

// ----------------------------------------------------------- 

IoObject *IoWeakLink_setLink(IoObject *self, IoObject *locals, IoMessage *m)
{ 
	/*#io
	docSlot("setLink(aValue)", 
		   "Sets the link pointer. Returns self.")
	*/
	IoObject *v = IoMessage_locals_valueArgAt_(m, locals, 0);
	IoWeakLink_justSetLink(self, v);
	return self; 
}

void IoWeakLink_justSetLink(IoObject *self, IoObject *v)
{
	IoState_removeWeakLink_(IOSTATE, self);
	IoWeakLink_unlink(self);
	DATA(self)->link = v; /* no IOREF because this is a weak link */
	
	if (!ISNIL(v)) 
	{
		v->mark.hasWeakLink = 1;
	}
	
	IoState_addWeakLink_(IOSTATE, self);
}

IoObject *IoWeakLink_link(IoObject *self, IoObject *locals, IoMessage *m)
{ 
	/*#io
	docSlot("link", 
		   "Returns the link pointer or Nil if none is set.")
	*/
	
	return DATA(self)->link; 
}

IoObject *IoWeakLink_setDelegate(IoObject *self, IoObject *locals, IoMessage *m)
{
	/*#io
	docSlot("setDelegate(aValue)", 
		   "Sets the delegate. Returns self. ")
	*/ 
	
	IoObject *v = IoMessage_locals_valueArgAt_(m, locals, 0);
	DATA(self)->delegate = IOREF(v);
	return self;
}

IoObject *IoWeakLink_delegate(IoObject *self, IoObject *locals, IoMessage *m)
{ 
	/*#io
	docSlot("delegate", 
		   "Returns the delegate or Nil if none is set.")
	*/
	
	return DATA(self)->delegate; 
}

IoObject *IoWeakLink_setIdentifier(IoObject *self, IoObject *locals, IoMessage *m)
{
	/*#io
	docSlot("setIdentifier(aValue)", 
		   "Sets the identifier. Returns self.")
	*/ 
	
	IoObject *v = IoMessage_locals_valueArgAt_(m, locals, 0);
	DATA(self)->identifier = IOREF(v);
	return self;
}

IoObject *IoWeakLink_identifier(IoObject *self, IoObject *locals, IoMessage *m)
{ 
	/*#io
	docSlot("identifier", 
		   "Returns the identifier or Nil if none is set.")
	*/
	
	return DATA(self)->identifier; 
}
