#ifdef WIN32
/* Enable fibers */
#define _WIN32_WINNT 0x0400
#include <windows.h>
#endif

#include "Base.h" 
#include "Coro.h" 
#include "Scheduler.h" 
#include <stdlib.h>
#include <string.h>
#include "PortableStrlcpy.h"

Coro *Coro_new(void)
{
	Coro *self = (Coro *)calloc(1, sizeof(Coro));
#ifdef HAS_FIBERS
	self->fiber = NULL;
#else
	self->stackSize = CSTACK_STACKSIZE;
	self->stack = calloc(1, self->stackSize + 16);
#endif
	Coro_loop(self);
	Coro_label_(self, "");
	return self;
}

void Coro_free(Coro *self)
{
#ifdef HAS_FIBERS
	// If this coro has a fiber, delete it.
	// Don't delete the main fiber. We don't want to commit suicide.
	if (self->fiber && self != Scheduler_mainCoro(self->scheduler)) {
		DeleteFiber(self->fiber);
	}
#endif
	Coro_remove(self);
	
	if (self->label) 
	{
		free(self->label);
	}
	
	memset(self, 0x0, sizeof(Coro)); // for debugging
	
#ifndef HAS_FIBERS
	free(self->stack);
#endif
	free(self);
}

void Coro_freeCoroChain(Coro *self)
{
	Coro *coro = self->next;
	
	while (coro != self) 
	{ 
		Coro *current = coro;
		coro = coro->next;
		Coro_free(current); 
	}
	
	Coro_free(self);
}

void Coro_scheduler_(Coro *self, void *scheduler)
{
	self->scheduler = scheduler;
}

void *Coro_scheduler(Coro *self)
{
	return self->scheduler;
}

void *Coro_stack(Coro *self)
{
#ifdef HAS_FIBERS
	return NULL;
#else
	return self->stack;
#endif
}

size_t Coro_stackSize(Coro *self)
{
#ifdef HAS_FIBERS
	return 0x0;
#else
	return self->stackSize;
#endif
}

void Coro_label_(Coro *self, const char *label)
{ 
	size_t len = strlen(label);
	char *ptr = (char *)realloc(self->label, len + 1);
	PortableStrlcpy(ptr, label, len);
	self->label = ptr;
}

const char *Coro_label(Coro *self) 
{ 
	return self->label; 
}

void Coro_loop(Coro *self)
{
	self->next = self;
	self->prev = self;
}

void Coro_remove(Coro *self)
{
	self->next->prev = self->prev;
	self->prev->next = self->next;
	Coro_loop(self);
}

void Coro_check(Coro *self)
{
	if (!self->next)
	{
		printf("Coro %p missing next\n", (void *)self);
		exit(1);
	}
	
	if (!self->prev)
	{
		printf("Coro %p missing prev\n", (void *)self);
		exit(1);
	}
}

void Coro_insertAfter_(Coro *self, Coro *other)
{
	Coro_remove(self);
	
	self->prev = other;
	self->next = other->next;
	
	other->next->prev = self;
	other->next = self;
	
	Coro_check(self);
}

void Coro_insertBefore_(Coro *self, Coro *other)
{
	Coro_remove(self);
	
	self->prev = other->prev;
	self->next = other;
	
	other->prev->next = self;
	other->prev = self;
	
	Coro_check(self);
}

unsigned int Coro_hasCoro_(Coro *self, Coro *other)
{
	Coro *c = self;
	
	do
	{ 
		if (c == other) 
		{
			return 1; 
		}
		
		c = c->next;
	} while (c != self);
	
	return 0;
}

void Coro_shouldPause_(Coro *self, unsigned char b) 
{
	self->shouldPause = (b!=0); 
}

unsigned int Coro_shouldPause(Coro *self) 
{
	return self->shouldPause; 
}

void Coro_shouldKill_(Coro *self, unsigned char b) 
{ 
	self->shouldKill = (b != 0); 
}

unsigned int Coro_shouldKill(Coro *self) 
{ 
	return self->shouldKill; 
}

inline unsigned char *Coro_endOfStack(Coro *self)
{ 
#ifdef HAS_FIBERS
	return NULL;
#else
	return ((unsigned char *)self->stack) + self->stackSize; 
#endif
}

size_t Coro_bytesLeftOnStack(Coro *self)
{
#ifdef HAS_FIBERS
	return 0;
#else
	unsigned char dummy;
	return (&dummy) - Coro_endOfStack(self);
#endif
}

// callback 

void Coro_object_(Coro *self, void *object) 
{ 
	self->object = object; 
}

void *Coro_object(Coro *self) 
{ 
	return self->object; 
}

void Coro_unyieldCallback_(Coro *self, CoroUnyieldCallback *callback)
{
	self->unyieldCallback = callback;
}

CoroUnyieldCallback *Coro_unyieldCallback(Coro *self)
{
	return self->unyieldCallback;
}

void Coro_doUnyieldCallback(Coro *self)
{
	if (!self->object)
	{
		 printf("Coro error: no object set on coro %p\n", (void *)self);
		 //exit(1);
	}
	 
	if (self->unyieldCallback)
	{
		self->unyieldCallback(self->object);
	}
	/*
	 else
	 {
		 printf("Coro error: no unyield callback on coro %p\n", (void *)self);
		 //exit(1);
	 }
	 */
}

// actions

void Coro_yield(Coro *self)
{
	Scheduler_yield((Scheduler *)self->scheduler);
}

void Coro_prepareToPause(Coro *self)
{
	Scheduler_prepareToPauseCoro_((Scheduler *)self->scheduler, self);
}

void Coro_pause(Coro *self)
{
	Coro_prepareToPause(self);
	Coro_yield(self);
}

void Coro_resume(Coro *self)
{
	Scheduler_resumeCoro_((Scheduler *)self->scheduler, self);
}

void Coro_kill(Coro *self)
{
	Scheduler_prepareToKillCoro_((Scheduler *)self->scheduler, self);
}

// status 

int Coro_isCurrent(Coro *self)
{
	return Scheduler_currentCoro((Scheduler *)self->scheduler) == self;
}

int Coro_isPaused(Coro *self)
{
	return Scheduler_hasPausedCoro_((Scheduler *)self->scheduler, self);
}

void Coro_show(Coro *self)
{
	if (self == 0x0)
	{
		printf("NULL");
	}
	else
	{
		printf("Coro_%p %s", (void *)self, self->label);
	}
}

