/*   
 *   copyright: Steve Dekorte, 2002. All rights reserved.
 *   license: See _BSDLicense.txt.
 */
 
#define MARKEDSTACK_C
#include "MarkedStack.h" 
#undef MARKEDSTACK_C

MarkedStack *MarkedStack_new(void)
{
  MarkedStack *self = (MarkedStack *)calloc(1, sizeof(MarkedStack));
  self->maxCount = MARKEDSTACK_START_SIZE;
  self->items = calloc(1, self->maxCount * sizeof(StackItem));
  self->count = 0;
  return self;
}

void MarkedStack_free(MarkedStack *self)
{
  free(self->items);
  free(self);
}

MarkedStack *MarkedStack_clone(MarkedStack *self)
{
  MarkedStack *s = MarkedStack_new();
  MarkedStack_copy_(s, self);
  return s;
}

void MarkedStack_sizeTo_(MarkedStack *self, int index)
{
  if (index >= self->maxCount) 
  {
    int newCount = self->maxCount * LIST_RESIZE_FACTOR;
    if (index >= newCount) { newCount = index + 1; }
    self->items = realloc(self->items, newCount * sizeof(StackItem)); 
    memset(self->items + self->count, 0x0, ((newCount - self->count) * sizeof(StackItem)));
    self->maxCount = newCount;
  }
}

void MarkedStack_copy_(MarkedStack *self, MarkedStack *other)
{
  MarkedStack_sizeTo_(self, other->maxCount);
  memcpy(self->items, other->items, other->count * sizeof(StackItem));
  self->maxCount = other->maxCount;
  self->count = other->count;
}

/* --- stack -------------------------------------------------- */

size_t MarkedStack_memorySize(MarkedStack *self)
{ return sizeof(MarkedStack) + (self->maxCount * sizeof(StackItem)); }

void MarkedStack_compact(MarkedStack *self) 
{
  size_t size = self->count * sizeof(StackItem);
  self->items = realloc(self->items, size);
}

void MarkedStack_resize(MarkedStack *self)
{
  int oldSize = (self->memEnd - self->items)*sizeof(StackItem);
  int newSize = oldSize*MARKEDSTACK_RESIZE_FACTOR;
  int i = self->top - self->items;
  self->items = realloc(self->items, newSize);
  self->top = self->items + i;
  self->memEnd = self->items + (newSize/sizeof(StackItem));
}

void MarkedStack_clear(MarkedStack *self) { self->top = self->items; }
size_t MarkedStack_totalSize(MarkedStack *self) { return (self->top - self->items); }

/* --- sizing ------------------------------------------------ */

void MarkedStack_do_on_(MarkedStack *self, MarkedStackDoOnCallback *callback, void *target)
{
  MarkedStack *stack = MarkedStack_newCopyWithoutMarks(self);
  int i;
  for(i = 0; i < MarkedStack_count(stack); i++)
  { (*callback)(target, MarkedStack_at_(stack, i)); }
  MarkedStack_free(stack);
}

MarkedStack *MarkedStack_newCopyWithoutMarks(MarkedStack *self)
{
  MarkedStack *newMarkedStack = MarkedStack_new();
  void **itemP = self->top;
  ptrdiff_t mark = self->lastMark;
  while (itemP > self->items)
  { 
    if (itemP - self->items == mark) 
    { mark = (ptrdiff_t)(*itemP); } 
    else 
    { MarkedStack_push_(newMarkedStack,*itemP); }
    itemP --;
  }
  return newMarkedStack;
}

void MarkedStack_popToMark_(MarkedStack *self, ptrdiff_t mark)
{
  while (self->lastMark && self->lastMark != mark)
  { MarkedStack_popMark(self); }
  if (self->lastMark == 0)
  {
      printf("MarkedStack error: unable to find mark %i in %p\n", mark, (void *)self);
      exit(1); 
  }
  MarkedStack_popMark(self);
}

